@@ -35,13 +35,13 @@
 #include <string>
 #include <vector>
 
+#include "enchant-provider.h"
 #include "sp_spell.h"
 #include "ispell_checker.h"
 #include "enchant.h"
-#include "enchant-provider.h"
 
 #ifndef ENCHANT_ISPELL_HOME_DIR
-#define ENCHANT_ISPELL_HOME_DIR ".enchant", "ispell"
+#define ENCHANT_ISPELL_HOME_DIR "ispell"
 #endif
 
 ENCHANT_PLUGIN_DECLARE("Ispell")
@@ -52,108 +52,45 @@ ENCHANT_PLUGIN_DECLARE("Ispell")
 
 typedef struct str_ispell_map
 {
-	char * lang;
-	char * dict;
-	char * enc;
+	const char * lang;
+	const char * dict;
+	const char * enc;
 } IspellMap;
 
 static const IspellMap ispell_map [] = {
 	{"ca"    ,"catala.hash"         ,"iso-8859-1" },
-	{"ca_ES" ,"catala.hash"         ,"iso-8859-1" },
 	{"cs"    ,"czech.hash"          ,"iso-8859-2" },
-	{"cs_CZ" ,"czech.hash"          ,"iso-8859-2" },
 	{"da"    ,"dansk.hash"          ,"iso-8859-1" },
-	{"da_DK" ,"dansk.hash"          ,"iso-8859-1" },
 	{"de"    ,"deutsch.hash"        ,"iso-8859-1" },
 	{"de_CH" ,"swiss.hash"          ,"iso-8859-1" },
-	{"de_AT" ,"deutsch.hash"        ,"iso-8859-1" },
-	{"de_DE" ,"deutsch.hash"        ,"iso-8859-1" },
 	{"el"    ,"ellhnika.hash"       ,"iso-8859-7" },
-	{"el_GR" ,"ellhnika.hash"       ,"iso-8859-7" },
 	{"en"    ,"british.hash"        ,"iso-8859-1" },
-	{"en_AU" ,"british.hash"        ,"iso-8859-1" },
-	{"en_BZ" ,"british.hash"        ,"iso-8859-1" },
-	{"en_CA" ,"british.hash"        ,"iso-8859-1" },
-	{"en_GB" ,"british.hash"        ,"iso-8859-1" },
-	{"en_IE" ,"british.hash"        ,"iso-8859-1" },
-	{"en_JM" ,"british.hash"        ,"iso-8859-1" },
-	{"en_NZ" ,"british.hash"        ,"iso-8859-1" },
-	{"en_TT" ,"british.hash"        ,"iso-8859-1" },
-	{"en_ZA" ,"british.hash"        ,"iso-8859-1" },
-	{"en_ZW" ,"british.hash"        ,"iso-8859-1" },
 	{"en_PH" ,"american.hash"       ,"iso-8859-1" },
 	{"en_US" ,"american.hash"       ,"iso-8859-1" },
 	{"eo"    ,"esperanto.hash"      ,"iso-8859-3" },
 	{"es"    ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_AR" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_BO" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_CL" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_CO" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_CR" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_DO" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_EC" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_ES" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_GT" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_HN" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_MX" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_NI" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_PA" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_PE" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_PR" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_PY" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_SV" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_UY" ,"espanol.hash"        ,"iso-8859-1" },
-	{"es_VE" ,"espanol.hash"        ,"iso-8859-1" },
 	{"fi"    ,"finnish.hash"        ,"iso-8859-1" },
-	{"fi_FI" ,"finnish.hash"        ,"iso-8859-1" },
 	{"fr"    ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_BE" ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_CA" ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_CH" ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_FR" ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_LU" ,"francais.hash"       ,"iso-8859-1" },
-	{"fr_MC" ,"francais.hash"       ,"iso-8859-1" },
 	{"hu"    ,"hungarian.hash"      ,"iso-8859-2" },
-	{"hu_HU" ,"hungarian.hash"      ,"iso-8859-2" },
 	{"ga"    ,"irish.hash"          ,"iso-8859-1" },
-	{"ga_IE" ,"irish.hash"          ,"iso-8859-1" },
 	{"gl"    ,"galician.hash"       ,"iso-8859-1" },
-	{"gl_ES" ,"galician.hash"       ,"iso-8859-1" },
 	{"ia"    ,"interlingua.hash"    ,"iso-8859-1" },
 	{"it"    ,"italian.hash"        ,"iso-8859-1" },
-	{"it_IT" ,"italian.hash"        ,"iso-8859-1" },
-	{"it_CH" ,"italian.hash"        ,"iso-8859-1" },
 	{"la"    ,"mlatin.hash"         ,"iso-8859-1" },
-	{"la_IT" ,"mlatin.hash"         ,"iso-8859-1" },
 	{"lt"    ,"lietuviu.hash"       ,"iso-8859-13" },
-	{"lt_LT" ,"lietuviu.hash"       ,"iso-8859-13" },
 	{"nl"    ,"nederlands.hash"     ,"iso-8859-1" },
-	{"nl_NL" ,"nederlands.hash"     ,"iso-8859-1" },
-	{"nl_BE" ,"nederlands.hash"     ,"iso-8859-1" },
 	{"nb"    ,"norsk.hash"          ,"iso-8859-1" },
-	{"nb_NO" ,"norsk.hash"          ,"iso-8859-1" },
 	{"nn"    ,"nynorsk.hash"        ,"iso-8859-1" },
-	{"nn_NO" ,"nynorsk.hash"        ,"iso-8859-1" },
 	{"no"    ,"norsk.hash"          ,"iso-8859-1" },
-	{"no_NO" ,"norsk.hash"          ,"iso-8859-1" },
 	{"pl"    ,"polish.hash"         ,"iso-8859-2" },
-	{"pl_PL" ,"polish.hash"         ,"iso-8859-2" },
 	{"pt"    ,"brazilian.hash"      ,"iso-8859-1" },
-	{"pt_BR" ,"brazilian.hash"      ,"iso-8859-1" },
 	{"pt_PT" ,"portugues.hash"      ,"iso-8859-1" },
 	{"ru"    ,"russian.hash"        ,"koi8-r" },
-	{"ru_MD" ,"russian.hash"        ,"koi8-r" },
-	{"ru_RU" ,"russian.hash"        ,"koi8-r" },
 	{"sc"    ,"sardinian.hash"      ,"iso-8859-1" },
-	{"sc_IT" ,"sardinian.hash"      ,"iso-8859-1" },
 	{"sk"    ,"slovak.hash"         ,"iso-8859-2" },
-	{"sk_SK" ,"slovak.hash"         ,"iso-8859-2" },
 	{"sl"    ,"slovensko.hash"      ,"iso-8859-2" },
-	{"sl_SI" ,"slovensko.hash"      ,"iso-8859-2" },
 	{"sv"    ,"svenska.hash"        ,"iso-8859-1" },
-	{"sv_SE" ,"svenska.hash"        ,"iso-8859-1" },
 	{"uk"    ,"ukrainian.hash"      ,"koi8-u" },
-	{"uk_UA" ,"ukrainian.hash"      ,"koi8-u" },
 	{"yi"    ,"yiddish-yivo.hash"   ,"utf-8" }
 };
 
@@ -253,13 +190,17 @@ ISpellChecker::checkWord(const char * const utf8Word, size_t length)
 	else
 		{
 			/* convert to 8bit string and null terminate */
-			size_t len_in, len_out;
-			char *In = (char *)(utf8Word);
+			size_t len_in, len_out, result;
+			// the 8bit encodings use precomposed forms
+			char *normalizedWord = g_utf8_normalize (utf8Word, length, G_NORMALIZE_NFC);
+			char *In = normalizedWord;
 			char *Out = szWord;
 			
-			len_in = length * sizeof(char);
+			len_in = strlen(In);
 			len_out = sizeof( szWord ) - 1;
-			if ((size_t)-1 == g_iconv(m_translate_in, &In, &len_in, &Out, &len_out))
+			result = g_iconv(m_translate_in, &In, &len_in, &Out, &len_out);
+			g_free(normalizedWord);
+			if ((size_t)-1 == result)
 				return false;
 			*Out = '\0';
 		}
@@ -297,12 +238,16 @@ ISpellChecker::suggestWord(const char * const utf8Word, size_t length,
 		{
 			/* convert to 8bit string and null terminate */
 			
-			size_t len_in, len_out;
-			char *In = (char *)(utf8Word);
+			size_t len_in, len_out, result;
+			// the 8bit encodings use precomposed forms
+			char *normalizedWord = g_utf8_normalize (utf8Word, length, G_NORMALIZE_NFC);
+			char *In = normalizedWord;
 			char *Out = word8;
-			len_in = length;
+			len_in = strlen(In);
 			len_out = sizeof( word8 ) - 1;
-			if ((size_t)-1 == g_iconv(m_translate_in, &In, &len_in, &Out, &len_out))
+			result = g_iconv(m_translate_in, &In, &len_in, &Out, &len_out);
+			g_free(normalizedWord);
+			if ((size_t)-1 == result)
 				return NULL;
 			*Out = '\0';
 		}
@@ -355,9 +300,26 @@ ISpellChecker::suggestWord(const char * const utf8Word, size_t length,
 	return sugg_arr;
 }
 
-static char *
-ispell_checker_get_prefix (void)
+static GSList *
+ispell_checker_get_dictionary_dirs (void)
 {
+	GSList *dirs = NULL;
+
+	{
+		GSList *config_dirs, *iter;
+
+		config_dirs = enchant_get_user_config_dirs ();
+		
+		for (iter = config_dirs; iter; iter = iter->next)
+			{
+				dirs = g_slist_append (dirs, g_build_filename ((const gchar *)iter->data, 
+									       ENCHANT_ISPELL_HOME_DIR, NULL));
+			}
+
+		g_slist_foreach (config_dirs, (GFunc)g_free, NULL);
+		g_slist_free (config_dirs);
+	}
+
 	/* until I work out how to link the modules against enchant in MacOSX - fjf
 	 */
 #ifndef XP_TARGET_COCOA
@@ -366,7 +328,7 @@ ispell_checker_get_prefix (void)
 	/* Look for explicitly set registry values */
 	ispell_prefix = enchant_get_registry_value ("Ispell", "Data_Dir");
 	if (ispell_prefix)
-		return ispell_prefix;
+		dirs = g_slist_append (dirs, ispell_prefix);
 
 	/* Dynamically locate library and search for modules relative to it. */
 	char * enchant_prefix = enchant_get_prefix_dir();
@@ -374,49 +336,36 @@ ispell_checker_get_prefix (void)
 		{
 			ispell_prefix = g_build_filename(enchant_prefix, "share", "enchant", "ispell", NULL);
 			g_free(enchant_prefix);
-			return ispell_prefix;
+			dirs = g_slist_append (dirs, ispell_prefix);
 		}
 #endif
 
 #ifdef ENCHANT_ISPELL_DICT_DIR
-	return g_strdup (ENCHANT_ISPELL_DICT_DIR);
-#else
-	return NULL;
+	dirs = g_slist_append (dirs, g_strdup (ENCHANT_ISPELL_DICT_DIR));
 #endif
+
+	return dirs;
 }
 
 static void
 s_buildHashNames (std::vector<std::string> & names, const char * dict)
 {
-	char * tmp, * private_dir, * home_dir, * ispell_prefix;
-
 	names.clear ();
 
-	/* until I work out how to link the modules against enchant in MacOSX - fjf
-	 */
-#ifndef XP_TARGET_COCOA
-	home_dir = enchant_get_user_home_dir ();
-#else
-	home_dir = g_strdup (g_getenv ("HOME"));
-#endif
-	if (home_dir) {
-		private_dir = g_build_filename (home_dir, ENCHANT_ISPELL_HOME_DIR, NULL);
-		
-		tmp = g_build_filename (private_dir, dict, NULL);
-		names.push_back (tmp);
-		g_free (tmp);
+	GSList *dirs, *iter;
 
-		g_free (private_dir);
-		g_free (home_dir);
-	}
+	dirs = ispell_checker_get_dictionary_dirs();
+	for (iter = dirs; iter; iter = iter->next)
+		{
+			char *tmp;
 
-	ispell_prefix = ispell_checker_get_prefix ();
-	if (ispell_prefix) {
-		tmp = g_build_filename (ispell_prefix, dict, NULL);
-		names.push_back (tmp);
-		g_free (tmp);
-		g_free (ispell_prefix);
-	}
+			tmp = g_build_filename ((const gchar *)iter->data, dict, NULL);
+			names.push_back (tmp);
+			g_free (tmp);
+		}
+
+	g_slist_foreach (dirs, (GFunc)g_free, NULL);
+	g_slist_free (dirs);
 }
 
 char *
@@ -446,8 +395,8 @@ ISpellChecker::loadDictionaryForLanguage ( const char * szLang )
 {
 	char *hashname = NULL;
 	
-	char * encoding = NULL;
-	char * szFile = NULL;
+	const char * encoding = NULL;
+	const char * szFile = NULL;
 	
 	for (size_t i = 0; i < size_ispell_map; i++)
 		{
@@ -701,13 +650,13 @@ ispell_provider_dispose (EnchantProvider * me)
 	g_free (me);
 }
 
-static char *
+static const char *
 ispell_provider_identify (EnchantProvider * me)
 {
 	return "ispell";
 }
 
-static char *
+static const char *
 ispell_provider_describe (EnchantProvider * me)
 {
 	return "Ispell Provider";
@@ -716,6 +665,9 @@ ispell_provider_describe (EnchantProvider * me)
 extern "C" {
 
 ENCHANT_MODULE_EXPORT (EnchantProvider *) 
+	     init_enchant_provider (void);
+
+EnchantProvider *
 init_enchant_provider (void)
 {
 	EnchantProvider *provider;
